//------------------------------------------------------------------------------
//
//  npgldraw.c
//  ANTZ application based on NPE
//
//  Created by Shane Saxon on 2/19/10.
//  Copyright Saxon Digital 2010. All rights reserved.
//
//------------------------------------------------------------------------------

#include "npgldraw.h"

#include <stdlib.h>

#include "../npglut.h"
#include "../../data/npplot.h"
#include "../../npctrl.h"


#define kTorusScreenSizeCount 4
#define kTorusThicknessCount 4

static GLint pin_display_list;
static GLint pin_torus_dl;

void DrawDefault (NPplotPtr plot);
void DrawCamera (NPplotPtr plot);
void DrawSurface (NPplotPtr plot);
void DrawPoints (NPplotPtr plot);
void DrawPin (NPplotPtr plot);
void DrawVideo (NPplotPtr plot);
void DrawGrid (NPplotPtr plot);


struct NPtorus {
	GLfloatXYZPtr	vArray;
	GLint			vArraySize;
	GLint*			quadArray;
	GLint			quadArraySize;
	int				slices;
	int				segments;
	GLfloat			thickness;
};
typedef struct NPtorus NPtorus;
typedef struct NPtorus * NPtorusPtr;

struct NPcircle {
	GLfloatXYPtr	vArray;
	int				segments;
	GLfloat			radius;
};
typedef struct NPcircle NPcircle;
typedef struct NPcircle * NPcirclePtr;


void InitTorus (void* dataRef);

GLuint CreatePinDL();
void DrawPinDL();

GLuint CreateTorusDL();
void DrawTorusDL();

void npDrawTorus (GLfloat radius, GLfloat thickness);

void CreateTorus (NPtorusPtr torus);
void CreateCircle (NPcirclePtr circle);
void DeleteCircle (NPcirclePtr circle);



//------------------------------------------------------------------------------
void npInitGLDraw(void* dataRef)
{
//	InitTorus (dataRef);

	pin_display_list = CreatePinDL();
	pin_torus_dl = CreateTorusDL();
}


// creates a set of toroids and loads them into GPU memory
//------------------------------------------------------------------------------
void InitTorus (void* dataRef)
{
	int i = 0, j = 0;
	NPtorusPtr torus = NULL;

	torus = (NPtorusPtr) malloc (sizeof(NPtorus));


	// 5x5, 7x5, 12x6, 16x6, 24x8, 32x10, 48x12, 64x16, 96x24, 128x32, 196x48
	for ( i = 0; i < kTorusScreenSizeCount; i++)
		for ( j = 0; j < kTorusThicknessCount; j++)
		{
			switch (i)
			{
				case 0 : torus->slices = 5; torus->segments = 5; break;
				case 1 : torus->slices = 5; torus->segments = 7; break;
				case 2 : torus->slices = 6; torus->segments = 12; break;
				case 3 : torus->slices = 6; torus->segments = 16; break;
				case 4 : torus->slices = 8; torus->segments = 24; break;
				case 5 : torus->slices = 10; torus->segments = 32; break;
				case 6 : torus->slices = 12; torus->segments = 48; break;
				case 7 : torus->slices = 16; torus->segments = 64; break;
				case 8 : torus->slices = 24; torus->segments = 96; break;
				case 9 : torus->slices = 32; torus->segments = 128; break;
				case 10 : torus->slices = 48; torus->segments = 196; break;
				case 11 : torus->slices = 64; torus->segments = 256; break;
			}
			
			torus->thickness =	(GLfloat) j * 100.0f / 
								(GLfloat) (kTorusThicknessCount - 1);

			CreateTorus (torus);
			// load torus to GPU and deallocate memory
			// we could save time by allocating enough memory for all torus's at once
			// or enough for largest torus then de-allocating all at once
		}
}


// creates a single torus of specified slices, stacks and thickness
//------------------------------------------------------------------------------
void CreateTorus (NPtorusPtr torus)
{
/*	int i=0, j=0;

	NPcirclePtr ring = (NPcirclePtr) malloc (sizeof(NPcircle));
	NPcirclePtr hoops = (NPcirclePtr) malloc (torus->slices * sizeof(NPcircle));

	torus->vArraySize = torus->slices * torus->segments;
	torus->vArray = (GLfloatXYZPtr) malloc (torus->vArraySize * sizeof(GLfloatXYZ));

	torus->quadArraySize = torus->vArraySize * 4;
	torus->quadArray = (GLint*) malloc (torus->quadArraySize * sizeof(GLint));

	// create our tube ring
	ring->radius = torus->thickness;		// 1.0 is no hole in the donut
	ring->segments = torus->slices;
	CreateCircle (ring);

	// create our hoops, the length segments of the tube
	for (i=0; i < torus->slices; i++)
	{
		hoops[i].radius = 1.0f + ring->vArray[i].x;
		hoops[i].segments = torus->segments;
		CreateCircle (&hoops[i]);
	}

	// build the vertex array for the torus
	for (i=0; i < torus->segments; i++)
		for (j=0; j < torus->slices; j++)
		{
			torus->vArray[i*torus->slices + j].x = hoops[j].vArray[i].x;		//the hoop provides x and y
			torus->vArray[i*torus->slices + j].y = hoops[j].vArray[i].y;
			torus->vArray[i*torus->slices + j].z = ring->vArray[j].y;		//the ring provides the z
		}

	// build the the torus polygon array
	for (i=0; i < torus->segments; i++)
		for (j=0; j < torus->slices; j++)
		{	
			if ( i == (torus->segments - 1) )		// the last torus segment
			{
				if ( j == (torus->slices - 1) )		// the last polygon
				{
					torus->quadArray[i][j][0] = (i * torus->slices) + j;
					torus->quadArray[i][j][1] = (i * torus->slices);
					torus->quadArray[i][j][2] = 0;
					torus->quadArray[i][j][3] = j;
				}
				else
				{
					torus->quadArray[i][j][0] = (i * torus->slices) + j;
					torus->quadArray[i][j][1] = (i * torus->slices) + (j + 1);
					torus->quadArray[i][j][2] = (j + 1);
					torus->quadArray[i][j][3] = j;
				}
			}
			else
			{
				if ( j == (torus->slices - 1) )		// the last polygon in a segment
				{
					torus->quadArray[i][j][0] = (i * torus->slices) + j;
					torus->quadArray[i][j][1] = (i * torus->slices);
					torus->quadArray[i][j][2] = ((i + 1) * torus->slices);
					torus->quadArray[i][j][3] = ((i + 1) * torus->slices) + j;
				}
				else
				{
					torus->quadArray[i][j][0] = (i * torus->slices) + j;
					torus->quadArray[i][j][1] = (i * torus->slices) + (j + 1);
					torus->quadArray[i][j][2] = ((i + 1) * torus->slices) + (j + 1);
					torus->quadArray[i][j][3] = ((i + 1) * torus->slices) + j;
				}
			}

	DeleteCircle (ring);
	for (i=0; i < torus->slices; i++)
		DeleteCircle (&hoops[i]);
*/
}


// creates a circle of specified number of points and radius
// centered on the origin (0,0) starting at 12 o'clock going CW
// really should be part of GL, glCircle!, lol
//------------------------------------------------------------------------------
void CreateCircle (NPcirclePtr circle)
{
/*	int i = 0;
	GLfloat angle = 0.0, deltaAngle = 0.0f;

	circle->vArray = (GLfloatXY) malloc (cirlce->segments * sizeof(GLfloatXY));


	if (circle->segments)
		deltaAngle = (2.0 * kPI) / circle->segments;
	
	// go around the circle, 
	for (angle = 0.0f; angle <= 1.99999f * kPI; angle += deltaAngle)
	{
//		circle->vArray[i]->x = circle->radius * (GLfloat)sin(angle);
//		circle->vArray[i].y = circle->radius * (GLfloat)cos(angle);
		i++;
	}
*/
}


//------------------------------------------------------------------------------
void DeleteCircle (NPcirclePtr circle)
{
	free (circle->vArray);
	free (circle);
}


// draws a torus of specified size and thickness, stacks and slices determined
// by size in pixels on screen calculated by screen size and z distance
// radius from center of the donut to the center of the tube, thickness 0-100%
//------------------------------------------------------------------------------
void npDrawTorus (GLfloat radius, GLfloat thickness)
{
	int torusID = 0;

	// torusID = screenSize + (thickness / numberOfSizes);

	glPushMatrix();
		glScalef(radius, radius, radius);
		glCallList(torusID);
	glPopMatrix();
}



//------------------------------------------------------------------------------
void DrawPinDL() 
{
// Draw Pin
	glTranslatef (0.0f, 0.0f, 5.0f);
	glRotatef (180.0f, 0.0f, 1.0f, 0.0f);
	npGLSolidCone (0.5f, 5.0f, 10, 1);

	glRotatef (-180.0f, 0.0f, 1.0f, 0.0f);	//undo rotation n save on pushMatrix
	npGLSolidSphere (0.5f, 10, 10);
}

//------------------------------------------------------------------------------
void DrawTorusDL() 
{
	// Draw Torus
	npGLWireTorus (0.15f, 1.5f, 7, 16);
}



//------------------------------------------------------------------------------
GLuint CreatePinDL()
{
	GLuint pinDL;

	// Create the id for the list
	pinDL = glGenLists(1);

	// start list
	glNewList (pinDL,GL_COMPILE);

	// call the function that contains the rendering commands
	DrawPinDL();

	// endList
	glEndList();

	return (pinDL);
}

//------------------------------------------------------------------------------
GLuint CreateTorusDL()
{
	GLuint pinDL;

	// Create the id for the list
	pinDL = glGenLists(2);

	// start list
	glNewList (pinDL,GL_COMPILE);

	// call the function that contains the rendering commands
	DrawTorusDL();

	// endList
	glEndList();

	return (pinDL);
}

								
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
void npDrawPlots (void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;
	NPplotPtr plot = NULL;

	for (i=0; i < data->map.plotCount; i++)
	{
		plot = data->map.plot[i];

		if (!plot->hide)			// don't draw hidden plots
			switch (plot->type)
			{
				case kPlotDefault : DrawDefault (plot); break;
				case kPlotCamera : DrawCamera (plot); break;
				case kPlotSurface : DrawSurface (plot); break;
				case kPlotPoints : DrawPoints (plot); break;
				case kPlotPin : DrawPin (plot); break;
				case kPlotVideo : DrawVideo (plot); break;
				case kPlotGrid : DrawGrid (plot); break;
			}
	}
}


//------------------------------------------------------------------------------
void DrawDefault (NPplotPtr plot)
{

}


// draw an icon at the location of the camera in the scene
//------------------------------------------------------------------------------
void DrawCamera (NPplotPtr plot)
{
	// draw icon of camera in the scene, but not the current viewport camera
}


//------------------------------------------------------------------------------
void DrawSurface (NPplotPtr plot)
{
	// draw ground
	glPushMatrix();

	// position the plot
	glTranslatef (plot->translate.x, plot->translate.y, plot->translate.z);
	glRotatef (plot->rotate.angle, plot->rotate.x, plot->rotate.y, plot->rotate.z);
	glScalef (plot->scale.x, plot->scale.y, plot->scale.z);

	glColor4ub (plot->color.r, plot->color.b, plot->color.g, plot->color.a);

	glBegin(GL_QUADS);
		glVertex3f (-100.0f, -100.0f, 0.0f);
		glVertex3f (-100.0f,  100.0f, 0.0f);
		glVertex3f ( 100.0f,  100.0f, 0.0f);
		glVertex3f ( 100.0f, -100.0f, 0.0f);
	glEnd();

	glPopMatrix();
}


//------------------------------------------------------------------------------
void DrawPoints (NPplotPtr plot)
{

}


//------------------------------------------------------------------------------
void DrawPin (NPplotPtr plot)
{
	int i = 0;

	NPplotPtr plotChild = NULL;

	glPushMatrix();

	// position the plot
	glTranslatef (plot->translate.x, plot->translate.y, plot->translate.z);
	glScalef (plot->scale.x, plot->scale.y, plot->scale.z);
	glRotatef (plot->rotate.angle, plot->rotate.x, plot->rotate.y, plot->rotate.z);

	glColor4ub (plot->color.r, plot->color.b, plot->color.g, plot->color.a);
//	glColor3ub (222, 0, 0);


	// first draw child here, then break it up, draw plots should iterate down, debug, zz
	// the tree branches and take care of selected glows, etc...
	// this routine should be generic regardless of branch level
	
	if (plot->surfaceType == kSurfacePin)
			glCallList (pin_display_list);

	//should call DrawPlot, that calls draw pin, move this to drawplots...   debug, zz
	for (i=0; i < plot->childCount; i++)
	{
		glPushMatrix();

		plotChild = plot->child[i];

		if (plotChild->surfaceType == kSurfaceTorus)
		{
//			glTranslatef (	plotChild->translate.x, 
//							plotChild->translate.y, 
//							plotChild->translate.z );
			glScalef (plotChild->scale.x, plotChild->scale.y, plotChild->scale.z);
			glRotatef (plotChild->rotate.angle, plotChild->rotate.x, plotChild->rotate.y, plotChild->rotate.z);

//			glRotatef (90.0f, 1.0f, 0.0f, 0.0f);

				if (plotChild->branchLevel % 2)		//is odd
						{
					
						}
						else								//is even
						{
							glRotatef (90.0f, 1.0f, 0.0f, 0.0f);
	/*						plotChild->rotate.angle = 35.0f;
							plotChild->rotate.x = 0.0f;
							plotChild->rotate.y = 1.0f;
							plotChild->rotate.z = 1.0f;
	*/
						}


			glCallList (pin_torus_dl);

			if (plotChild->lineWidth > 0.0f)
			{
				//draw a wireframe slightly scaled up on top...
			}
			if (plotChild->pointSize > 0.0f)
			{
				//draw points slightly scaled up on top...
			}
		}
		glPopMatrix();
	}

//	npGLWireTorus(0.15f, 1.5f, 7, 16);

	glPopMatrix();
}


//------------------------------------------------------------------------------
void DrawVideo (NPplotPtr plot)
{

}


//------------------------------------------------------------------------------
void DrawGrid (NPplotPtr plot)
{
	int			i = 0;
	int			j = 0;
	int			k = 0;
	GLfloat		x = 0.0f;
	GLfloat		y = 0.0f;
	GLfloat		z = 0.0f;
	GLfloatXYZ	length;

	NPgridPtr grid = (NPgridPtr) plot->data;

	length.x = grid->size.x * grid->segments.x;
	length.y = grid->size.y * grid->segments.y;
	length.z = grid->size.z * grid->segments.z;

	glPushMatrix();

	// position the plot
	glTranslatef (plot->translate.x, plot->translate.y, plot->translate.z);
	glRotatef (plot->rotate.angle, plot->rotate.x, plot->rotate.y, plot->rotate.z);
	glScalef (plot->scale.x, plot->scale.y, plot->scale.z);

	glTranslatef (length.x / -2.0f, length.y / -2.0f, 0.0f);

	glColor4ub (plot->color.r, plot->color.b, plot->color.g, plot->color.a);

//	glDisable (GL_LIGHTING);

	glBegin (GL_LINES);

		glLineWidth (plot->lineWidth);

		for (i=0; i <= grid->segments.z; i++)
		{
			for (j=0; j <= grid->segments.x; j++)
			{
				glVertex3f (x, 0.0f, z);
				glVertex3f (x, length.y, z);
				glVertex3f (0.0f, y, z);
				glVertex3f (length.x, y, z);
				x += grid->size.x;
				y += grid->size.y;
			}
			x = 0.0f;
			y = 0.0f;
			z += grid->size.z;
		}
	glEnd();

//	glEnable (GL_LIGHTING);

	glPopMatrix();
}


// draw 2D overlays including grids, text, messages, highlight objects, etc...
//------------------------------------------------------------------------------
void npDrawHUD (void* dataRef)
{

}


/////////////////////////////updates a single plot with new data if not frozen and draws it if not hidden
void SetIndexColor (GLubyteRGBA *color, int *colorIndex)
{	
	int index = *colorIndex;
	
	GLubyte colorPallet[kColorIndexSize][4] = {	
											{255, 255, 255, 255},
											{0, 255, 0, 255},
											{255, 230, 0, 255},
											{0, 50, 255, 255},
											{255, 0, 0, 255},
											{152, 0, 255, 255},
											{255, 168, 0, 255},
											{0, 255, 255, 255},
											{255, 0, 255, 255},
											{0, 153, 0, 255},
											{173, 255, 202, 255},
											{255, 180, 255, 255},
											{0, 152, 255, 255},
											{185, 255, 0, 255},
											{152, 0, 0, 255},
											{50, 101, 101, 255},
											{127, 127, 255, 255},
											{185, 153, 102, 255},
											{197, 82, 0, 255},
											{127, 127, 127, 255}  };
											
	//make sure colorIndex is inbounds, if not, then put it inbounds																			
	if (index >= kColorIndexSize)
		while (index >= kColorIndexSize)
			index -= kColorIndexSize;		//subtracts pallete sized increments
	
	if (index < 0)
		while (index < 0)
			index += kColorIndexSize;		//adds pallete sized increments

	//now set the color
	color->r = colorPallet[index][0];
	color->g = colorPallet[index][1];
	color->b = colorPallet[index][2];
	color->a = colorPallet[index][3];		//perhaps better not to change alpha, debug

	*colorIndex = index;
}

